#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXLINELEN 1024
typedef enum { false, true } bool;

/*-------------------------------------------------------------------------
 * Miscellaneous functions.
 *-------------------------------------------------------------------------*/

char ungetline_text[MAXLINELEN];
bool ungetline_valid = false;
int lineno = 0;

char *getline(char *buf)
{
	if (ungetline_valid) {
		ungetline_valid = false;
		strcpy(buf, ungetline_text);
		return buf;
	}

	if (fgets(buf, MAXLINELEN, stdin) == NULL)
		return NULL;

	++lineno;
	buf[strlen(buf)-1] = '\0';	/* Strip newline. */
	return buf;
}

void ungetline(char *s)
{
	strcpy(ungetline_text, s);
	ungetline_valid = true;
}

char *filter(char *s)
{
	static char buf[MAXLINELEN];
	char *p = buf;
	while (*s != '\0') {
		if (*s != '_')
			*p++ = *s;
		else {
			*p++ = '\\';
			*p++ = '_';
		}
		++s;
	}
	*p = '\0';
	return buf;
}

void filter_block(char *buf, bool ignore_at);

void filter_alltt(char *buf, bool open)
{
	if (!open)
		printf("\\begin{alltt}\n");
	filter_block(buf, false);
	printf("\\end{alltt}\n");
}

void filter_verbatim(char *buf, bool open)
{
	if (!open)
		printf("\\begin{verbatim}\n");
	while (getline(buf) != NULL) {
		if (buf[0] == '@') {
			ungetline(buf);
			break;
		}
		printf("%s\n", buf);
	}
	printf("\\end{verbatim}\n");
}

void filter_block(char *buf, bool ignore_at)
{
	char *p;

	while (getline(buf) != NULL) {
		if (buf[0] == '@')
			switch (buf[1]) {
			case 'V':
				filter_verbatim(buf, 0);
				continue;
			case '@':
				if (!ignore_at)
					return;
				continue;
			case 'L':
				printf("%s\n", &buf[3]);
				continue;
			case 'H':
				continue;
			case 't':
			case 'i':
			case 'b':
				break;
			default:
				ungetline(buf);
				return;
			}
		for (p = buf; *p != '\0'; ++p)
			if (*p == '@') {
				int t = *++p;
				switch (t) {
				case 't': printf("\\verb$"); break;
				case 'i': printf("\\textnit{"); break;
				case 'b': printf("\\textnbf{"); break;
				default: fprintf(stderr, "%d: Invalid type `%c'\n", lineno, t);
				}
				++p;
				while (*++p != '}')
					putchar(*p);
				if (t == 't')
					putchar('$');
				else
					putchar('}');
			} else if (*p == '#')
				printf("\\#");
			else
				putchar(*p);
		putchar('\n');
	}
}

void filter_headers(char *buf)
{
	if (*buf != '\0') {
		char *s;
		s = strtok(buf, " ");
		while (s != NULL) {
			printf("#include <%s>\n", s);
			s = strtok(NULL, " ");
		}
		putchar('\n');
	}
}

/*-------------------------------------------------------------------------
 * Parsing functions.
 *-------------------------------------------------------------------------*/

char header_name[MAXLINELEN];

void do_xref(char *buf)
{
	printf("\\Xref{%s}{%s}{sym:%s}\n",
	       header_name, filter(buf), buf);
}

void do_prototype(char *buf)
{
	printf("\\item[Prototype]\n"
	       "\\begin{alltt}\n");
	filter_headers(buf);
	filter_alltt(buf, true);
}

void do_description(char *buf)
{
	printf("\\item[Description]\n");
	filter_block(buf, true);
}

void do_implementation(char *buf)
{
	printf("\\item[Implementation]\n"
	       "\\begin{verbatim}\n");
	filter_headers(&buf[2]);
	filter_verbatim(buf, true);
}

void do_example(char *buf)
{
	printf("\\item[Example]\n"
	       "\\begin{verbatim}\n");
	filter_headers(&buf[2]);
	filter_verbatim(buf, true);
}

static bool done_1990=false, done_1999=false;

void do_iso1990_ref(char *buf)
{
	if (done_1999)
		printf(", \\isooldref{%s}", buf);
	else
		printf("\\item[References] \\isooldref{%s}", buf);
	done_1990 = true;
}

void do_iso1999_ref(char *buf)
{
	if (done_1990)
		printf(", \\isonewref{%s}", buf);
	else
		printf("\\item[References] \\isonewref{%s}", buf);
	done_1999 = true;
}

void do_entry_command(char *buf)
{
	switch (buf[0]) {
	case 'p':
		do_prototype(&buf[2]);
		break;
	case 'd':
		do_description(buf);
		break;
	case 'm':
		do_implementation(buf);
		break;
	case 'e':
		do_example(buf);
		break;
	case 'r':
		do_iso1990_ref(&buf[2]);
		break;
	case 'R':
		do_iso1999_ref(&buf[2]);
		break;
	case 'L':
		printf("%s\n", &buf[2]);
		break;
	case 'H':
		break;
	default:
		fprintf(stderr, "%d: Invalid command `%c'\n", lineno, buf[0]);
	}
}

void do_entry(char *buf)
{
	static bool first_entry = true;
	int i;
	if (!first_entry) {
		printf("%% ");
		for (i = 0; i < 77; ++i)
			putchar('-');
		putchar('\n');
	} else
		first_entry = false;
	done_1990 = done_1999 = false;

	printf("\\subsection{\\texttt{%s}}\n", filter(buf));
	printf("\\label{sym:%s}\n", buf);
	printf("\\bfvindex{%s}{%s}\n", header_name, filter(buf));
	printf("\\begin{basedescript}{\\desclabelwidth{.5cm}\\desclabelstyle{\\nextlinelabel}}\n");

	while (getline(buf) != NULL)
		if (buf[0] == '@') {
			if (buf[1] == ':' || buf[1] == 'x') {
				ungetline(buf);
				break;
			} else
				do_entry_command(&buf[1]);
		} else if (buf[0] != '\0')
			fprintf(stderr, "%d: Invalid line `%s'\n", lineno, buf);

	printf("\n\\end{basedescript}\n");
}

void do_command(char *buf)
{
	switch (buf[0]) {
	case 'h':
		strcpy(header_name, &buf[2]);
		printf("%% Header `%s'\n\n", header_name);
		break;
	case 'x':
		do_xref(&buf[2]);
		break;
	case ':':
		do_entry(&buf[2]);
		break;
	case 'L':
		printf("%s\n", &buf[2]);
		break;
	case 'H':
		break;
	default:
		fprintf(stderr, "%d: Invalid command `%c'\n", lineno, buf[0]);
	}
}

int main(void)
{
	char buf[MAXLINELEN];

	while (getline(buf) != NULL)
		if (buf[0] == '@')
			do_command(&buf[1]);

	return 0;
}
